#include "rumd/IHS_OutputManager.h"
#include "rumd/Sample.h"
#include "rumd/Potential.h"
#include "rumd/IntegratorIHS.h"

IHS_OutputManager::IHS_OutputManager(Sample* S, float timeStep) : LogLinOutputManager(S, S->GetOutputDirectory(), "ihs"), max_num_iter(1000), num_iter_inner(50), stoppingCriterion(1.e-4f), gz_ihs_energyFile(0), writeInterval(0), has_logLin(true) {
  
  ihsSample = new Sample(*sample);
  ihsSample->SetVerbose(false);
  itg = new IntegratorIHS(timeStep);
 
  std::vector<Potential*>* potentialList = sample->GetPotentials();
  std::vector<Potential*>::iterator potIter;
  for(potIter = potentialList->begin(); potIter != potentialList->end(); potIter++) {
    Potential* pot_copy = (*potIter)->Copy();
    ihs_potentialList.push_back( pot_copy ); // so can delete them later
    ihsSample->AddPotential( pot_copy );
  }
  
  
  ihsSample->SetIntegrator(itg);
  
  ihsSample->SetWriteEnergies(true);
  ihsSample->GetEnergiesMetaData().Set("potentialEnergy", true);
  ihsSample->GetEnergiesMetaData().Set("virial", true);
  ihsSample->GetEnergiesMetaData().Set("kineticEnergy", false);
  ihsSample->GetEnergiesMetaData().Set("totalEnergy", false);
  ihsSample->GetEnergiesMetaData().Set("thermostat_Ps", false);
  ihsSample->GetEnergiesMetaData().Set("totalEnergy", false);
  ihsSample->GetEnergiesMetaData().Set("pressure", false);

  
  ihsWriter = new ConfigurationWriterReader();
  ihsWriter->metaData.precision = 4;
  ihsWriter->metaData.has_images = true;
  ihsWriter->metaData.has_velocities = false;
  ihsWriter->metaData.has_forces = false;
  ihsWriter->metaData.has_pot_energies = false;
  ihsWriter->metaData.has_virials = false;
  ihsWriter->metaData.has_logLin = true;  

  SetIHSLogLinParams(1,0 );
}


IHS_OutputManager::~IHS_OutputManager() {
  delete ihsSample;
  delete itg;
  delete ihsWriter;
  
  std::vector<Potential*>::iterator potIter;
  for(potIter = ihs_potentialList.begin(); potIter != ihs_potentialList.end(); potIter++)
    delete (*potIter);
  
}



void IHS_OutputManager::SetIHSLinearWriteInterval(unsigned long int writeInterval) { 
  SetLogLinParams(writeInterval, writeInterval);
  this->writeInterval = writeInterval;
  has_logLin = false;
  SetDuplicateBlockEnds(false);
}


void IHS_OutputManager::SetIHSLogLinParams(unsigned int base, unsigned int maxInterval) {
  writeInterval = 0;
  has_logLin = true;
  SetLogLinParams(base, maxInterval);
  SetDuplicateBlockEnds(true);
}


void IHS_OutputManager::Write() {
  
  // need to avoid doing the minimization on the same configuration
  // twice when duplicateBlockEnds is true

  (*ihsSample) = (*sample); // copies the data
  unsigned int num_total_iter = 0;

  // calc max force
  float max_force2 = 0.0f;
  float sc2 = stoppingCriterion*stoppingCriterion;
  ihsSample->CalcF();
  ihsSample->GetParticleData().CopyForFromDevice();
  
  for(unsigned int atIdx = 0;atIdx < ihsSample->GetNumberOfParticles(); atIdx++)
    {
      float4& f = ihsSample->GetParticleData().h_f[atIdx];
      float f2 = f.x*f.x + f.y*f.y + f.z*f.z;
      if( f2 > max_force2 )
	max_force2 = f2;
    }
  while (max_force2 > sc2 && num_total_iter < max_num_iter) {
    for(unsigned  int j=0; j < num_iter_inner; j++ ){
      itg->Integrate();
      ihsSample->CalcF();
    }
    ihsSample->GetParticleData().CopyForFromDevice();
    max_force2 = 0.0f;
    for(unsigned int atIdx = 0; atIdx < ihsSample->GetNumberOfParticles(); atIdx++)
      {
	float4& f = ihsSample->GetParticleData().h_f[atIdx];
	float f2 = f.x*f.x + f.y*f.y + f.z*f.z;
	if( f2 > max_force2 )
	  max_force2 = f2;
      }    
    num_total_iter += num_iter_inner;
  }
  

  std::ostringstream conf_filename;
  conf_filename << outputDirectory << "/" << baseFilename << std::setfill('0') << std::setw(numFilenameDigits) << std::setprecision(numFilenameDigits) << logLin.block << ".xyz";
  
  // set the logLin details in writer->metaData (this would not necessarily make sense for all Configuration writers)  
  
  ihsWriter->metaData.logLin = logLin;
  ihsWriter->SetVerbose(verbose);

  if(logLin.index > 0) 
    ihsWriter->Write(ihsSample, conf_filename.str(), "a");
  else
    ihsWriter->Write(ihsSample, conf_filename.str(), "w");
  

  

 if(logLin.index == 0) // (start a new file)
    {
      if(gz_ihs_energyFile) gzclose(gz_ihs_energyFile);
      std::ostringstream ihs_energyFilename;
      ihs_energyFilename << outputDirectory << "/ihsEnergies" << std::setfill('0') << std::setw(numFilenameDigits) << std::setprecision(numFilenameDigits) << logLin.block << ".dat.gz";
      gz_ihs_energyFile = gzopen(ihs_energyFilename.str().c_str(),"w");
      
      // comment line at top of new file
      std::map<std::string, bool>::const_iterator it;
      float dt = 0.;
      if(sample->GetIntegrator())
	dt = sample->GetIntegrator()->GetTimeStep();
      
      
      gzprintf(gz_ihs_energyFile, "# N=%d",ihsSample->GetNumberOfParticles());

      if(has_logLin)
	gzprintf(gz_ihs_energyFile," timeStepIndex=%u logLin=%u,%u,%u,%u,%u", logLin.nextTimeStep, logLin.block, logLin.base, logLin.index, logLin.maxIndex, logLin.maxInterval);
      else
	gzprintf(gz_ihs_energyFile," Dt=%f", dt*writeInterval);

      gzprintf(gz_ihs_energyFile," columns=pe,virial,max_force,num_iter\n");
    } // end if(logLin.index == 0)  
 
  int precision = 6;
  gzprintf(gz_ihs_energyFile, "%.*g ", precision, ihsSample->GetPotentialEnergy()/ihsSample->GetNumberOfParticles());
  gzprintf(gz_ihs_energyFile, "%.*g ", precision, ihsSample->GetVirial()/ihsSample->GetNumberOfParticles());
  gzprintf(gz_ihs_energyFile, "%.*g ", precision, sqrt(max_force2));
  gzprintf(gz_ihs_energyFile, "%u ", num_total_iter);

  gzprintf(gz_ihs_energyFile, "\n");  

}

