#include "rumd/MoleculeData.h"

#include <cstdlib>
#include <cmath>
#include <iostream>

#include "rumd/RUMD_Error.h"

///////////////////////////////////////////////////////
// Contructor / Destructor
///////////////////////////////////////////////////////

MoleculeData::MoleculeData(Sample *globalS, const char *fin){
  // The Sample class
  S = globalS;
 
  // The simulation box class
  SB = dynamic_cast<RectangularSimulationBox*>(S->GetSimulationBox());

  // Set flags
  flagHarmonic = 0;
  flagFENE = 0;
  flagConstraint = 0;

  // Default max number of bonds
  max_num_bonds = 4;

  // Get number of molecules and max nuau from top file
  GetInfoFromTopology(fin);

  // Simply set the number of tpb to pbb (nice and cl comment!!)
  threads_per_block.x = S->GetParticlesPerBlock();
  num_blocks.x = S->GetNumberOfVirtualParticles()/threads_per_block.x;
 
  // Length of lists
  lblist = S->GetNumberOfVirtualParticles() + max_num_bonds*S->GetNumberOfParticles();
  lmlist = (max_num_uau + 1)*num_mol;
  lbtlist = max_num_bonds*S->GetNumberOfParticles();

  // blist: Allocate memory on host
  size_t nbytes = sizeof(uint1)*lblist;
  if ( cudaMallocHost((void **)&h_blist, nbytes) == cudaErrorMemoryAllocation )
    throw RUMD_Error("MoleculeData","MoleculeData","Memory allocation failure");
  
  // blist: Allocate memory on device
  if ( cudaMalloc((void **)&d_blist, nbytes) == cudaErrorMemoryAllocation)
    throw RUMD_Error("MoleculeData","MoleculeData","Memory allocation failure");
    
  // mlist: Allocate memory on host
  nbytes = sizeof(int1)*lmlist;
  if ( cudaMallocHost((void **)&h_mlist, nbytes) == cudaErrorMemoryAllocation)
    throw RUMD_Error("MoleculeData","MoleculeData","Memory allocation failure");
    
  // mlist: Allocate memory on the device
  if ( cudaMalloc((void **)&d_mlist, nbytes) == cudaErrorMemoryAllocation)
    throw RUMD_Error("MoleculeData","MoleculeData","Memory allocation failure");
 
  //btlist: On host   
  nbytes = sizeof(uint1)*lbtlist;
  if ( cudaMallocHost((void **)&h_btlist, nbytes) == cudaErrorMemoryAllocation)
    throw RUMD_Error("MoleculeData","MoleculeData","Memory allocation failure");

  // btlist: On device
  if ( cudaMalloc((void **)&d_btlist, nbytes) == cudaErrorMemoryAllocation)
    throw RUMD_Error("MoleculeData","MoleculeData","Memory allocation failure");
  
  // bplist: On host
  nbytes = sizeof(float2)*num_btypes;
  if ( cudaMallocHost((void **)&h_bplist, nbytes) == cudaErrorMemoryAllocation)
    throw RUMD_Error("MoleculeData","MoleculeData","Memory allocation failure");
  
  // bplist: On device
  if ( cudaMalloc((void **)&d_bplist, nbytes) == cudaErrorMemoryAllocation)
    throw RUMD_Error("MoleculeData","MoleculeData","Memory allocation failure");
 
  // btlist_int: On host
  nbytes = sizeof(int1)*num_btypes;
  if ( cudaMallocHost((void **)&h_btlist_int, nbytes) == cudaErrorMemoryAllocation)
    throw RUMD_Error("MoleculeData","MoleculeData","Memory allocation failure");
  
  // btlist_int: On device
  if ( cudaMalloc((void **)&d_btlist_int, nbytes) == cudaErrorMemoryAllocation)
    throw RUMD_Error("MoleculeData","MoleculeData","Memory allocation failure");
  

  // Initialisation of blist (is important later on)
  for ( unsigned n=0; n<lblist; n++ ) h_blist[n].x = 0;
	
  // Initialisation of mlist (is important later on)
  for ( unsigned n=0; n<lmlist; n++ ) h_mlist[n].x = -1;

  // Initialisation of btlist_int (is important later on)
  for ( unsigned n=0; n<num_btypes; n++ ) h_btlist_int[n].x = -1;

  // Read in the top file
  ReadTopology(fin);
}

MoleculeData::~MoleculeData(){
  cudaFreeHost(h_mlist);
  cudaFree(d_mlist);
  cudaFreeHost(h_blist);
  cudaFree(d_blist);
  cudaFreeHost(h_btlist);
  cudaFree(d_btlist);
  cudaFreeHost(h_bplist);
  cudaFree(d_bplist);
  cudaFreeHost(h_btlist_int);
  cudaFree(d_btlist_int);
}

///////////////////////////////////////////////////////
// Methods.
///////////////////////////////////////////////////////

// Set molecule lists from top file
void MoleculeData::ReadTopology(const char *str){
  char line[256];
  const char section[] = {'[', ' ', 'b', 'o', 'n', 'd', 's', ' ' , ']', '\n', '\0'};
  unsigned ai, aj, moli, btype;
  
  unsigned *mlist_index = new unsigned [num_mol];
  for ( unsigned n=0; n<num_mol; n++ ) mlist_index[n] = 0;
  
  unsigned *btlist_index = new unsigned [lbtlist];
  for ( unsigned n=0; n<lbtlist; n++ ) btlist_index[n] = 0;
  

  FILE *top = fopen(str, "r");
  if ( top == NULL )
    throw RUMD_Error("MoleculeData","ReadTopology","Couldn't open topology file.");


  // Set the file pointer to the line after 
  // the [ bond ] section string
  do { 
    fgets(line, 256, top);
  } while ( strcmp(line, section) != 0 ); 
  
  // Skip next line (comments)
  fgets(line, 256, top);

  // Go through the pair list
  do {
    
    // Check that we have an entry
    fpos_t fpos;
    fgetpos(top, &fpos);    
    fgets(line, 256, top);
    if ( strlen(line) < 3 ) break; //magic number... :o/

    // If so reset pointer to line start and read again
    // but with format
    fsetpos(top, &fpos);
    if ( fscanf(top, "%u%u%u%u\n", &moli, &ai, &aj, &btype) != 4 ) 
      throw RUMD_Error("MoleculeData","ReadTopology",
			"Incorrect file format");

    CheckRead(moli, ai, aj);

    // Add entries to blist
    AddblistEntry(ai, aj);

    // Add entries to mlist
    AddmlistEntry(mlist_index, moli, ai, aj);

    // Add entries into btlist 
    AddbtlistEntry(btlist_index, ai, aj, btype);

  } while ( !feof(top) );
  
  fclose(top);
  
  // Copy the arrays on to the device (Checks are missing)
  size_t nbytes = sizeof(uint1)*lblist;
  cudaMemcpy(d_blist, h_blist, nbytes, cudaMemcpyHostToDevice);
  
  nbytes = sizeof(int1)*lmlist;
  cudaMemcpy(d_mlist, h_mlist, nbytes, cudaMemcpyHostToDevice);

  nbytes = sizeof(int1)*lbtlist;
  cudaMemcpy(d_btlist, h_btlist, nbytes, cudaMemcpyHostToDevice);

  // Freeing the mlist index array
  delete[] mlist_index;
  delete[] btlist_index;
}

// Helper function for top-reader
void MoleculeData::CheckRead(unsigned i, unsigned ai, unsigned aj){
  if ( i >= num_mol )
    throw RUMD_Error("MoleculeData","CheckRead", "Molecule index in .top file is larger than number of molecules.");
  // MODIFY THIS.
  if ( ai >= S->GetNumberOfParticles() || aj >= S->GetNumberOfParticles() ) 
    throw RUMD_Error("MoleculeData","CheckRead", "Particle index in .top file is larger than number of particles\n.");
}

// Read the top file - get number of molecules and number of maximum united atomic units
void MoleculeData::GetInfoFromTopology(const char *str){
  char line[256];
  const char section[] = {'[', ' ', 'b', 'o', 'n', 'd', 's', ' ' , ']', '\n', '\0'};
  unsigned ai, aj, btype;
  int moli;
 
  int num_part = S->GetNumberOfParticles();

  FILE *top = fopen(str, "r");
  if ( top == NULL )
    throw RUMD_Error("MoleculeData","ReadInfoFromTopology",
		     "Couldn't open topology file.");

  // Set the file pointer to the line after 
  // the [ bond ] section string
  do { 
    fgets(line, 256, top);
  } while ( strcmp(line, section) != 0 ); 
  
  // Skip next line (comments)
  fgets(line, 256, top);

  int nmol = 0; num_btypes=0;
  int *marray = (int*)malloc(0);
  // Go through the pair list
  do {
    
    // Check that we have an entry
    fpos_t fpos;
    fgetpos(top, &fpos);    
    fgets(line, 256, top);
    if ( strlen(line) < 3 ) break; //magic number... :o/

    // If so, reset pointer to line start and read again
    // but with format
    fsetpos(top, &fpos);
    if ( fscanf(top, "%u%u%u%u\n", &moli, &ai, &aj, &btype) !=4 )
      throw RUMD_Error("MoleculeData","ReadInfoFromTopology",
		       "Format not correct");

    if ( btype > num_btypes ) num_btypes++;
 
    if ( moli>nmol-1 ){
      nmol = moli+1;
      marray = (int*)realloc(marray, nmol*sizeof(int));
      if ( marray == NULL )
	 throw RUMD_Error("MoleculeData","ReadInfoFromTopology", "Couldn't allocate memory");
      marray[moli] = 0;
    }
    marray[moli]++; 
    
  } while ( !feof(top) );
  
  fclose(top);

  int maxnuau = 0; 
  for ( int n=0; n<nmol; n++ ){ //nmol
    if ( marray[n] > maxnuau ) maxnuau = marray[n];
  }
  
  maxnuau++;
  num_btypes ++;

  num_mol = nmol;
  max_num_uau = maxnuau; 

  std::cout << "Topology file "<< str << " found. \n";  
  std::cout << "Read "<< num_mol << " molecules with " 
	    << max_num_uau << " maximum united atomic units. " 
	    << std::endl;
  std::cout << "Read " << num_btypes << " bonding type(s)."<< std::endl;

  free(marray);
}

///////////////////////////////////////////////////////
// Add from topology
///////////////////////////////////////////////////////

void MoleculeData::AddbtlistEntry(unsigned *index, unsigned ai, unsigned aj, unsigned type){
  if ( index[ai]==max_num_bonds || index[aj]==max_num_bonds )
      throw RUMD_Error("MoleculeData","ReadTotology", "Number of bonds has exceeded the max_num_bonds");

  h_btlist[ai*max_num_bonds + index[ai]].x = type;
  h_btlist[aj*max_num_bonds + index[aj]].x = type;
  
  index[ai] = index[ai] + 1;
  index[aj] = index[aj] + 1;
}

// Helper function for top-reader
void MoleculeData::AddblistEntry(unsigned ai, unsigned aj){
  unsigned i = S->GetNumberOfVirtualParticles() + ai*max_num_bonds + h_blist[ai].x;  
  h_blist[i].x = aj;
  
  i = S->GetNumberOfVirtualParticles() + aj*max_num_bonds + h_blist[aj].x;  
  h_blist[i].x = ai;
  
  h_blist[ai].x++;
  h_blist[aj].x++;
}

// Helper function for top-reader
void MoleculeData::AddmlistEntry(unsigned *index, unsigned i, unsigned ai, unsigned aj){
  int flag[2] = {0};
  unsigned si = i*max_num_uau;
  
  for ( unsigned n=0; n<index[i]; n++ ){
    if ( h_mlist[si+n+num_mol].x == (signed)ai ) 
      flag[0] = 1;
    else if ( h_mlist[si+n+num_mol].x == (signed)aj ) 
      flag[1] = 1;
  }

  if ( flag[0] == 0 ){
    h_mlist[si+index[i]+num_mol].x = ai;
    index[i]++;
  }
  
  if ( flag[1] == 0 ){
    h_mlist[si+index[i]+num_mol].x = aj;
    index[i]++;
  }
   
  h_mlist[i].x = (signed)index[i];
}

///////////////////////////////////////////////////////
// Set methods
///////////////////////////////////////////////////////

void MoleculeData::SetBondHarmonic( unsigned type, float lbond, float ks ){
  if( flagHarmonic == 0 ){
    HarmBond = new BondHarmonic(this, this->S);
    S->AddPotential(HarmBond);
    flagHarmonic = 1;
  }
  
  if ( type>num_btypes )
    throw RUMD_Error("MoleculeData","SetBondHarmonic", "Type index too large\n.");
  
  h_bplist[type].x = lbond;
  h_bplist[type].y = ks;
  
  h_btlist_int[type].x = 0;
  
  size_t nbytes = sizeof(float2)*num_btypes;
  cudaMemcpy(d_bplist, h_bplist, nbytes, cudaMemcpyHostToDevice);
  
  nbytes = sizeof(int1)*num_btypes;
  cudaMemcpy(d_btlist_int, h_btlist_int, nbytes, cudaMemcpyHostToDevice);
}

void MoleculeData::SetBondConstraint( unsigned bondType, float bondLength ){
  if( bondType > num_btypes || bondLength <= 0 )
    throw RUMD_Error("MoleculeData", "SetBondConstraint", "Type index too large or bond-length is negative.\n");
  
  // Update internal host and device side data.
  h_btlist_int[bondType].x = 2; 
  h_bplist[bondType].x = bondLength; 
  h_bplist[bondType].y = 0; // Not used.
  
  cudaMemcpy( d_btlist_int, h_btlist_int, num_btypes * sizeof(int1), cudaMemcpyHostToDevice );  
  cudaMemcpy( d_bplist, h_bplist, num_btypes * sizeof(float2), cudaMemcpyHostToDevice );
  
  unsigned maxBonds = GetMaximumNumberOfBonds();
  
  // Count the number of constraints;
  unsigned count = 0;
  for( unsigned i = 0; i < S->GetNumberOfParticles(); i++){
    unsigned realBonds = h_blist[i].x;
    
    for( unsigned j = 0; j < realBonds; j++){
      unsigned userBondType = h_btlist[i*maxBonds+j].x;
      unsigned internalBondType = h_btlist_int[userBondType].x;
      
      if( userBondType == bondType && internalBondType == 2 )
	count++;
    }
  }
  
  // Subtract 1 degree of freedom for each constraint.
  S->SubtractDOFs( count / 2 );
  
  if( flagConstraint == 0 ){
    constraintPotential = new ConstraintPotential(this, this->S);
    S->AddPotential(constraintPotential);
    flagConstraint = 1;
  }
  
  constraintPotential->UpdateConstraintLayout();
}

void MoleculeData::SetBondFENE(unsigned type){
  if( flagFENE == 0 ){
    FENEBond = new BondFENE(this, this->S);
    S->AddPotential(FENEBond);
    flagFENE = 1;
  }
  
  if ( type>num_btypes )
    throw RUMD_Error("MoleculeData","SetBondHarmonic", "Type index too large\n.");

  h_btlist_int[type].x = 1;

  size_t nbytes = sizeof(int1)*num_btypes;
  cudaMemcpy(d_btlist_int, h_btlist_int, nbytes, cudaMemcpyHostToDevice);
}

void MoleculeData::SetExclusionBond(PairPotential* potential, unsigned etype){
  potential->SetExclusionBond(h_btlist, h_blist, S->GetNumberOfParticles(), S->GetNumberOfVirtualParticles(), max_num_bonds, etype);
}

void MoleculeData::SetExclusionMolecule(PairPotential *potential, unsigned molindex){
  potential->SetExclusionMolecule(h_mlist, molindex, max_num_uau);
}

void MoleculeData::SetExclusionMolecule(PairPotential *potential){
  for ( unsigned n=0; n<num_mol; n++ ) potential->SetExclusionMolecule(h_mlist, n, max_num_uau);
}

///////////////////////////////////////////////////////
// Debuggin methods
///////////////////////////////////////////////////////

// Print mlist and blist (for debugging mostly)
void MoleculeData::PrintLists(void){

  std::cout << "--- blist: Sec I ---" << std::endl;
  for ( unsigned i=0; i<S->GetNumberOfVirtualParticles(); i++ )
    std::cout << h_blist[i].x << std::endl;
  
  std::cout << "--- blist: Sec II ---" << std::endl;
  for ( unsigned i=S->GetNumberOfVirtualParticles(); i<lblist; i++ )
    std::cout << h_blist[i].x << std::endl;
	
  std::cout << "--- mlist: ---" << std::endl;
  for ( unsigned i=0; i<lmlist; i++)
    std::cout << h_mlist[i].x << std::endl;
}

// Print mlist and blist for i (for debugging mostly)
void MoleculeData::PrintLists(unsigned i){
  
  if(i>S->GetNumberOfParticles())
    std::cout << "index " << i << " too large (number of particles is " << S->GetNumberOfParticles() << ")" << std::endl; 
    else {
  std::cout << "Atom index " << i << " has " <<  h_blist[i].x 
	    << " bonds." << " Binding indices "; 
  for ( unsigned n=0; n<h_blist[i].x; n++ )
    std::cout << h_blist[S->GetNumberOfVirtualParticles() 
			 + i*max_num_bonds + n].x << " ";
  std::cout << std::endl;
  }
}

// Print mlist to file - convenient for post run data analysis
void MoleculeData::PrintMlist(){
  FILE *fout = fopen("mlist.top", "w");

  if ( fout == NULL )
    throw RUMD_Error("MoleculeData","PrintMlist",
		     "Couldn't open file for writing - something seriously wrong!");

  std::cout << "Writing mlist to mlist.top\n";

  for ( unsigned n=0; n<lmlist; n++ )
    fprintf(fout, "%d\n", h_mlist[n].x);

  fclose(fout);
}
