/*
    Copyright (C) 2010  Thomas Schrøder

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    LICENSE.txt file for license details.
*/

#include <iostream>

#include "rumd/rumd_utils.h"
#include "rumd/PairPotential.h"
#include "rumd/SimulationBox.h"
#include "rumd/NeighborList.h"

using namespace std;

///////////////////////////////////////////
// CONSTRUCTION/DESTRUCTION 
///////////////////////////////////////////

NeighborList::NeighborList(){
  // These need to be zero before Allocate functions are first called
  numBlocks = 0;
  numBlocksParticles = 0;
  numVirtualParticles = 0;
  NbList = 0;
  skin = 0.3f; // default value
  
  if ( cudaMalloc( (void**) &d_shell, sizeof(float) * MaxNumTypes * MaxNumTypes ) == cudaErrorMemoryAllocation )
    throw( RUMD_Error("NeighborList","NeighborList","Malloc failed on d_shell") );
  d_ctrl = 0; // for checking whether initialized
}

NeighborList::~NeighborList(){
  Free();
  cudaFree(d_shell);
}

///////////////////////////////////////////
// MEMORY ALLOCATION/DEALLOCATION 
///////////////////////////////////////////

void NeighborList::Initialize(ParticleData* pData, KernelPlan& kPlan){
  particleData = pData;
  numParticles = particleData->GetNumberOfParticles();
  kp = kPlan;
  numVirtualParticles = kp.num_virt_part;
  unsigned int nb_nvp = kp.num_blocks * numVirtualParticles;
  
  // Reinitializes the check for update and returns.
  if( numBlocks == kp.num_blocks && numBlocksParticles == nb_nvp ){
    md_initialize_nblist_check<<<kp.grid, kp.threads>>>( d_ctrl );
    return;
  }
  
  Free();
 
  // Allocate NeighborList data. CPU.
  
  if( cudaMallocHost( (void**) &h_ExclusionList, sizeof(NBTYPEv) * nb_nvp ) )
    throw( RUMD_Error("NeighborList","Initialize","Malloc failed on h_ExclusionNeighborList") );
  
  // Allocate NeighborList data. GPU.
  
  if( cudaMalloc( (void**) &NbList, sizeof(NBTYPEv) * nb_nvp ) == cudaErrorMemoryAllocation )
    throw( RUMD_Error("NeighborList","Initialize","Malloc failed on NbList") );

  if( cudaMalloc( (void**) &d_savedNeighborList, sizeof(NBTYPEv) * nb_nvp ) == cudaErrorMemoryAllocation )
    throw( RUMD_Error("NeighborList","Initialize","Malloc failed on NbList") );
  
  if( cudaMalloc( (void**) &d_ExclusionList, sizeof(NBTYPEv) * nb_nvp ) )
    throw( RUMD_Error("NeighborList","Initialize","Malloc failed on d_ExclusionNeighborList") );
  
  // For checking whether to update
  if( cudaMalloc( (void**) &d_last_r, numVirtualParticles * sizeof(float4) ) == cudaErrorMemoryAllocation )
    throw( RUMD_Error("NeighborList","Initialize","Malloc failed on d_last_r") );

  if( cudaMalloc( (void**) &d_ctrl, sizeof(int) ) == cudaErrorMemoryAllocation )
    throw( RUMD_Error("NeighborList","Initialize","Malloc failed on d_ctrl") );

  if( cudaMalloc( (void**) &cell_list, sizeof(int) * kp.num_blocks * kp.num_blocks ) == cudaErrorMemoryAllocation )
    throw( RUMD_Error("NeighborList","Initialize","Malloc failed on cell_list") );
  

  cudaMemset( d_last_r, 0, numVirtualParticles * sizeof(float4) );
  // Initialize the ExclusionList with only 1's.  
  for( unsigned int i=0; i < nb_nvp; i++ )
    h_ExclusionList[i].x = ~( 0x00000000 );
  
  cudaMemcpy( d_ExclusionList, h_ExclusionList, sizeof(NBTYPEv) * nb_nvp, cudaMemcpyHostToDevice );
  
  // set the check to non-zero to trigger an update the first time
  md_initialize_nblist_check<<<kp.grid, kp.threads>>>( d_ctrl );
  
  numBlocksParticles = nb_nvp;  
  numBlocks = kp.num_blocks;
  particlePerBlock = kp.threads.x;

}

void NeighborList::Free(){
  if(numBlocksParticles != 0) {
    cudaFree(d_last_r); // note allocated size is proportional to nvp 
    cudaFree(NbList);
  }
  
  if(numBlocks != 0){
    cudaFree(d_ctrl);
    cudaFree(cell_list);
  }
}

///////////////////////////////////////////
// Main methods.
///////////////////////////////////////////

void NeighborList::UpdateNBlist(SimulationBox* simBox){
  float4* d_r = particleData->d_r;
  float4* d_im = particleData->d_im;
  dim3 threads1(kp.threads.x, 1);

  RectangularSimulationBox* testRSB = dynamic_cast<RectangularSimulationBox*>(simBox);
  if(!testRSB) throw RUMD_Error("NeighborList", "UpdateNBlist", "Unrecognized Simulation box type");

  // check whether a re-build is necessary
  displacement_since_nb_build<<< kp.grid, threads1 >>>( numParticles, d_r, d_last_r, d_im, testRSB, testRSB->GetDevicePointer(), 0.25*skin*skin, d_ctrl );
  
  md_update_nblist<<< kp.grid, kp.threads, kp.shared_size >>>( numParticles, d_r, d_im, d_shell, testRSB, testRSB->GetDevicePointer(), 
							       NbList, d_ExclusionList, cell_list, d_ctrl, d_last_r );
}

void NeighborList::ResetNeighborList(){
  md_initialize_nblist_check<<<kp.grid, kp.threads>>>( d_ctrl );
}

void NeighborList::SaveNeighborList(){
  cudaMemcpy( d_savedNeighborList, NbList, numBlocksParticles * sizeof(NBTYPEv), cudaMemcpyDeviceToDevice );
}

void NeighborList::RestoreNeighborList(){
  cudaMemcpy( NbList, d_savedNeighborList, numBlocksParticles * sizeof(NBTYPEv), cudaMemcpyDeviceToDevice );
}

///////////////////////////////////////////
// Set methods
///////////////////////////////////////////

void NeighborList::SetCutoffs(PairPotential* potential){
  for (unsigned int i=0; i<MaxNumTypes; i++){
    for (unsigned int j=0; j<MaxNumTypes; j++){ 
      unsigned int k = i * MaxNumTypes + j;
      double Rcut = (double) potential->GetPotentialParameter(i,j,0);
      h_shell[k] = Rcut + skin;
    }
  }  
  
  cudaMemcpy( d_shell, h_shell, sizeof(float)*MaxNumTypes*MaxNumTypes, cudaMemcpyHostToDevice);
  if(d_ctrl) // if Initialize has been called
    ResetNeighborList();
}

void NeighborList::SetCutoffs(PairPotential* potential, float set_skin){ skin = set_skin; SetCutoffs(potential); }

void NeighborList::SetExclusionList( NBTYPEv* newList ){
  for( unsigned int i=0; i < numBlocksParticles; i++ )
    h_ExclusionList[i] = newList[i];
  
  cudaMemcpy( d_ExclusionList, h_ExclusionList, sizeof(NBTYPEv) * numBlocksParticles, cudaMemcpyHostToDevice );
}

void NeighborList::SetExclusionBond(int1 *h_btlist, uint1 *h_blist, unsigned num_part, unsigned num_virt_part, unsigned max_num_bonds, unsigned etype){
  
  unsigned ppb =  particlePerBlock;

  for ( unsigned i=0; i<num_part; i++ ){
    for ( unsigned m=0; m<h_blist[i].x; m++ ){
      
      int offset = i*max_num_bonds + m;
      
      if ( h_btlist[offset].x == (signed)etype ){
	unsigned j = h_blist[num_virt_part + offset].x;
		
	unsigned a = i/ppb;
	unsigned b = j/ppb;
	
	unsigned ofnblist_i = a*numBlocks*ppb;
	unsigned ofnblist_j = b*numBlocks*ppb;
	
	unsigned ofnblist = ofnblist_i+ b*ppb + (i%ppb);
	h_ExclusionList[ofnblist].x &= ~( 1 << (j%ppb) );
	
	ofnblist = ofnblist_j+ a*ppb + (j%ppb);
	h_ExclusionList[ofnblist].x &= ~( 1 << (i%ppb) );
      }
      
    }
  }

  cudaMemcpy( d_ExclusionList, h_ExclusionList, sizeof(NBTYPEv) * numBlocksParticles, cudaMemcpyHostToDevice );
}

// Exclude two given particle pair interactions based on type.
void NeighborList::SetExclusionType( unsigned* type, unsigned type0, unsigned type1 ){
  if( numParticles == 0 )
    throw( RUMD_Error("NeighborList","SetExclusionType","No particles have been read") );
  
  for(unsigned i=0; i < numParticles; i++){
    for(unsigned j=i; j < numParticles; j++){
      if( (type[i] == type0 && type[j] == type1) || (type[i] == type1 && type[j] == type0) ){
	unsigned ofnblist_i = ( i / particlePerBlock ) * numBlocks * particlePerBlock + ( j / particlePerBlock ) * particlePerBlock + ( i % particlePerBlock );
	unsigned ofnblist_j = ( j / particlePerBlock ) * numBlocks * particlePerBlock + ( i / particlePerBlock ) * particlePerBlock + ( j % particlePerBlock );
	
	h_ExclusionList[ofnblist_i].x &= ~( 1 << ( j % particlePerBlock ) );
	h_ExclusionList[ofnblist_j].x &= ~( 1 << ( i % particlePerBlock ) );
      }
    }
  }

  // Eliminate virtual particles
  for(unsigned i=0; i < numParticles; i++){
    for(unsigned j=numParticles; j < numVirtualParticles; j++){
      unsigned ofnblist_i = ( i / particlePerBlock ) * numBlocks * particlePerBlock + ( j / particlePerBlock ) * particlePerBlock + ( i % particlePerBlock );
      unsigned ofnblist_j = ( j / particlePerBlock ) * numBlocks * particlePerBlock + ( i / particlePerBlock ) * particlePerBlock + ( j % particlePerBlock );
      
      h_ExclusionList[ofnblist_i].x &= ~( 1 << ( j % particlePerBlock ) );
      h_ExclusionList[ofnblist_j].x &= ~( 1 << ( i % particlePerBlock ) );
    }
  }
  cudaMemcpy( d_ExclusionList, h_ExclusionList, sizeof(NBTYPEv) * numBlocksParticles, cudaMemcpyHostToDevice );
}

void NeighborList::SetExclusionMolecule(int1 *h_mlist, unsigned molindex, unsigned max_num_uau){
  
  unsigned ppb =  particlePerBlock;
  unsigned i = molindex*max_num_uau;

  for ( unsigned n=i; n<i+max_num_uau-1; n++ ){
    int ai = h_mlist[n].x; 
    if ( ai == - 1) break; 

    for ( unsigned m=n+1; m<max_num_uau; m++ ){
      int bi = h_mlist[m].x;
      if ( bi == -1) break;
		
      unsigned a = (unsigned)ai/ppb;
      unsigned b = (unsigned)bi/ppb;
	
      unsigned ofnblist_i = a*numBlocks*ppb;
      unsigned ofnblist_j = b*numBlocks*ppb;
      
      unsigned ofnblist = ofnblist_i+ b*ppb + (ai%ppb);
      h_ExclusionList[ofnblist].x &= ~( 1 << (bi%ppb) );
      
      ofnblist = ofnblist_j+ a*ppb + (bi%ppb);
      h_ExclusionList[ofnblist].x &= ~( 1 << (ai%ppb) );
    }
  }

  cudaMemcpy( d_ExclusionList, h_ExclusionList, sizeof(NBTYPEv) * numBlocksParticles, cudaMemcpyHostToDevice );
}

///////////////////////////////////////////////////
// Print information about the internal structure.
///////////////////////////////////////////////////

// 32 bit assumption on bit-pattern. The interaction matrix is layed out in a non-standard pattern.
void NeighborList::PrintExclusionList(){
  for(unsigned i=0; i < numParticles; i++){
    cout << endl << endl << "Particle " << i << ": ";
    const unsigned int neighborListOffset = ( i / kp.threads.x ) * kp.num_blocks * kp.threads.x + ( i % kp.threads.x );

    for( unsigned j=0; j < numBlocks; j++ ){
      unsigned value = h_ExclusionList[neighborListOffset + j*kp.threads.x].x;
    
      if( value == 0 ){
	cout << "|0|";
      } // 2^32 - 1.
      else if(value == 4294967295){
	cout << "|1|";
      }
      else{ // Expand only non-trivial patterns.
	cout << "|";
	for(int z=kp.threads.x-1; z >= 0; z--){
	  unsigned m = 1 << z;
	  
	  if( ( value & m ) == 0 )
	    cout << "0";
	  else
	    cout << "1";
	}
	cout << "|";
      }
    }
  }
  cout << endl << endl;
} 

void NeighborList::PrintCellList(){
  unsigned int nb = numBlocks;
  int* h_cell_list; 
  int EmptyCells = 0;
  
  h_cell_list = (int*) malloc( sizeof(int) * nb * nb );
  CUDA_SAFE_CALL( cudaMemcpy( h_cell_list, cell_list, sizeof(int)*nb*nb, cudaMemcpyDeviceToHost) );
  
  //printf("cell_list:\n"); 
  for (unsigned int i=0; i<nb; i++) {
    //printf("%d: ", i);
    for (unsigned j=0; j<nb; j++) {
      //printf("%d ", h_cell_list[i*nb + j]);
      if (h_cell_list[i*nb + j]<0) EmptyCells++;
    }
    //printf("\n");
  }
  //printf("\n");

  //printf("Empty cells: %d, Density of empty cells %f\n", EmptyCells, float(EmptyCells)/float(nb*nb));
  std::cout << float(EmptyCells)/float(nb*nb) << std::endl;

  free(h_cell_list);
}

///////////////////////////////////////////////
// Kernel Implementations
///////////////////////////////////////////////

__global__  void md_initialize_nblist_check( int* updateNbList ){ updateNbList[0] = NumBlocks; }

// When to update the neighborlist.
template <class S> __global__ void displacement_since_nb_build( unsigned numParticles, float4* r, float4* last_r, float4* image, 
								S* simBox, float* simBoxPointer, float skin2_over_4, int* updateNbList ){
  if( MyGP < numParticles ){
    float array[simulationBoxSize];
    simBox->loadMemory( array, simBoxPointer );
    
    // Distance moved since last update.
    float distanceMoved = (simBox->calculateDistance(r[MyGP], last_r[MyGP], array)).w;
    
    // Need update?
    if( distanceMoved > skin2_over_4 ){
      // The number of blocks MUST be the same as for calls to md_update_nblist!
      updateNbList[0] = NumBlocks;
      image[MyGP].w = 0.f;
    }
    else{
      // Consistency check performed when configuration IO. 
      image[MyGP].w = distanceMoved; 
    }
  }
}

// Build neighborlist.
template <class S> __device__ void update_nblist( unsigned int num_part, float4 my_r, float4* r, float* params, S* simBox, float* simBoxPointer, 
						  NBTYPEv* NbList, NBTYPEv* ExclusionList, int *cell_list ){
  
  extern __shared__ float4 s_r[];
  unsigned int* s_NbList = (unsigned int*) s_r;

  const unsigned int offset = MyT*PPerBlock;
  const unsigned int CellOffset = MyB*NumBlocks*PPerBlock; 
  unsigned int Pread;
  NBTYPE NbListCell;
  unsigned int CellListCount = 0;
  float dist2;
  
  // Loop over all cells
  for (unsigned int cell0 = 0; cell0 < NumBlocks; cell0+=TPerPart) {
    
    unsigned int cell = cell0 + MyT;
    
    NbListCell = 0;
    Pread = __umul24(cell,PPerBlock) + MyP;
    
    if (cell<NumBlocks) {
      if (Pread<num_part) {
	s_r[MyP+offset] = r[Pread]; 
      } else {
	s_r[MyP+offset].w = -1.0f; // Mark particle as being 'dummy'
      }
    }    
    
    // Make sure all shared data loaded before using
    __syncthreads();  
    
    // Check all interactions in this cell
    if (cell<NumBlocks){
      if( MyGP < num_part ){ // If virtual just write 0.
	for (unsigned int i = 0; i < PPerBlock; i++){
	  float4 sr = s_r[i+offset];
	  if (sr.w>-0.5f) {
	    dist2 = (simBox->calculateDistance(my_r, sr, simBoxPointer)).w;
	    
	    if ( dist2 <= (params[__float_as_int(sr.w)]*params[__float_as_int(sr.w)]) && dist2 >= 0.00001f ){
	      NbListCell += (1<<i); 
	    }
	  }
	}
	
	NbListCell &= ExclusionList[CellOffset + Pread].x;
      }
      
      // Write the nb-list for this cell to global memory
      NbList[CellOffset + Pread].x = NbListCell;
    }
    
    // Check if there are any interactions with this cell (borrowing shared memory)
    __syncthreads();
    if (MyP==0) s_NbList[MyT] = 0;
    __syncthreads();
    atomicOr(&(s_NbList[MyT]), NbListCell);	// Expensive?
    __syncthreads();
    
    if (MyP==0 && MyT==0) {
      for (unsigned j=0; j<TPerPart; j++) {
      	if (s_NbList[j]>0) {
	  cell_list[MyB*NumBlocks + CellListCount] = cell0 + j;
	  CellListCount++;	  
	}  
      }  	 
    } 
    __syncthreads();     
    
  }

  // Mark end of cell_list array by -1
  if (MyP==0 && MyT==0) {	 
    for (unsigned j=MyB*NumBlocks + CellListCount; j<(MyB + 1)*NumBlocks; j++) {
      cell_list[j] = -1;
    }
  }  
}

template <class S> __global__ void md_update_nblist( unsigned int num_part, float4* r, float4* im, float* params, 
						     S* simBox, float* simBoxPointer, NBTYPEv* NbList, NBTYPEv* ExclusionList, 
						     int* cell_list, int* updateNbList, float4* last_r ){
  if(updateNbList[0]){
    // Load the simulation box in local memory to avoid bank conflicts.
    float array[simulationBoxSize];
    simBox->loadMemory( array, simBoxPointer );
    
    // Read LJ-parameters into shared memory
    __shared__ float s_params[MaxNumTypes*MaxNumTypes]; 
    
    if (MyP<MaxNumTypes*MaxNumTypes && MyT==0){ 
      s_params[MyP] = params[MyP];
    }
    
    extern __shared__ float4 s_r[];
    
    float4 my_r;
    
    if (MyT==0) { 
      my_r = r[MyGP];
      s_r[MyP] = my_r;
    }
    
    __syncthreads();
    
    my_r = s_r[MyP];
    
    __syncthreads();
    
    // Update nb-list for this particle
    update_nblist( num_part, my_r, r, &s_params[MaxNumTypes*__float_as_int(my_r.w)], simBox, array, NbList, ExclusionList, cell_list );
    
    if(MyT==0){
      last_r[MyGP] = my_r;
      if (MyP==0) atomicSub(updateNbList,1);
    }
  }
}
